package org.processmining.activityclusterarraycreator.algorithms;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.deckfour.xes.classification.XEventClass;
import org.processmining.activityclusterarray.models.ActivityClusterArray;
import org.processmining.activityclusterarray.models.impl.ActivityClusterArrayFactory;
import org.processmining.activityclusterarraycreator.parameters.ConvertCausalActivityGraphToActivityClusterArrayParameters;
import org.processmining.causalactivitygraph.models.CausalActivityGraph;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.util.Pair;

public class ConvertCausalActivityGraphToActivityClusterArrayAlgorithm {

	public ActivityClusterArray apply(PluginContext context, final CausalActivityGraph graph,
			ConvertCausalActivityGraphToActivityClusterArrayParameters parameters) {
		ActivityClusterArray clusters = ActivityClusterArrayFactory.createActivityClusterArray();
		List<XEventClass> activities = graph.getActivities();
		clusters.init(graph.getLabel(), new HashSet<XEventClass>(activities));
		List<Pair<XEventClass, XEventClass>> causalities = new ArrayList<Pair<XEventClass, XEventClass>>(
				graph.getSetCausalities());
		/*
		 * Sort the edges from heavy to light. As a result, we
		 * will start clustering with the most heavy edges.
		 */
		Collections.sort(causalities, new Comparator<Pair<XEventClass, XEventClass>>() {
			public int compare(Pair<XEventClass, XEventClass> p1, Pair<XEventClass, XEventClass> p2) {
				Double d1 = graph.getCausality(p1.getFirst(), p1.getSecond());
				Double d2 = graph.getCausality(p2.getFirst(), p2.getSecond());
				return d2.compareTo(d1);
			}
		});
		while (!causalities.isEmpty()) {
			Pair<XEventClass, XEventClass> selectedCausality = causalities.iterator().next();
			causalities.remove(selectedCausality);
			Set<XEventClass> sources = new HashSet<XEventClass>();
			Set<XEventClass> targets = new HashSet<XEventClass>();
			Set<XEventClass> cluster = new HashSet<XEventClass>();
			sources.add(selectedCausality.getFirst());
			targets.add(selectedCausality.getSecond());
			cluster.addAll(sources);
			cluster.addAll(targets);
			boolean notReady = true;
			while (notReady && cluster.size() < parameters.getClusterThreshold()) {
				notReady = false;
				List<Pair<XEventClass, XEventClass>> newCausalities = new ArrayList<Pair<XEventClass, XEventClass>>();
				for (Pair<XEventClass, XEventClass> causality : causalities) {
					if (sources.contains(causality.getFirst()) || targets.contains(causality.getSecond())) {
						if (graph.getCausality(causality.getFirst(), causality.getSecond()) > parameters
								.getClusterFactor() * Math.log(1.0 * cluster.size())) {
							notReady = sources.add(causality.getFirst()) || notReady;
							notReady = targets.add(causality.getSecond()) || notReady;
							cluster.add(causality.getFirst());
							cluster.add(causality.getSecond());
						}
					} else if (parameters.isCheckBackwardArc() && targets.contains(causality.getFirst())
							&& sources.contains(causality.getSecond())) {
						if (graph.getCausality(causality.getFirst(), causality.getSecond()) > parameters
								.getClusterFactor() * Math.log(1.0 * cluster.size())) {
							notReady = sources.add(causality.getFirst()) || notReady;
							notReady = targets.add(causality.getSecond()) || notReady;
							cluster.add(causality.getFirst());
							cluster.add(causality.getSecond());
						}
					} else {
						newCausalities.add(causality);
					}
				}
				causalities = newCausalities;
			}
			//			System.out.println(cluster);
			clusters.addCluster(cluster);
			clusters.setInputs(cluster, sources);
			clusters.setOutputs(cluster, targets);
		}
		ActivityClusterArray sanitiziedClusters = ActivityClusterArrayFactory.createActivityClusterArray();
		sanitiziedClusters.init(clusters);
		Set<XEventClass> unClusteredActivities = new HashSet<XEventClass>();
		if (parameters.isIncludeAll()) {
			for (XEventClass activity : activities) {
				boolean found = false;
				for (Set<XEventClass> cluster : clusters.getClusters()) {
					if (cluster.contains(activity)) {
						found = true;
					}
				}
				if (!found) {
					unClusteredActivities.add(activity);
				}
			}
			if (!unClusteredActivities.isEmpty()) {
				sanitiziedClusters.addCluster(unClusteredActivities);
				sanitiziedClusters.setInputs(unClusteredActivities, new HashSet<XEventClass>());
				sanitiziedClusters.setOutputs(unClusteredActivities, new HashSet<XEventClass>());
			}
		}
		for (Set<XEventClass> cluster : clusters.getClusters()) {
			boolean remove = false;
			// Prevent subclusters.
			for (Set<XEventClass> otherCluster : clusters.getClusters()) {
				if ((cluster != otherCluster) && cluster.size() < otherCluster.size()
						&& otherCluster.containsAll(cluster)) {
					remove = true;
					Set<XEventClass> inputs = clusters.getInputs(otherCluster);
					inputs.addAll(clusters.getInputs(cluster));
					clusters.setInputs(otherCluster, inputs);
					Set<XEventClass> outputs = clusters.getOutputs(otherCluster);
					outputs.addAll(clusters.getOutputs(cluster));
					clusters.setOutputs(otherCluster, outputs);
				}
			}
			// Prevent identical clusters.
			for (Set<XEventClass> otherCluster : sanitiziedClusters.getClusters()) {
				if (cluster.equals(otherCluster) 
						&& clusters.getInputs(cluster).equals(sanitiziedClusters.getInputs(otherCluster)) 
						&& clusters.getOutputs(cluster).equals(sanitiziedClusters.getOutputs(otherCluster))) {
					remove = true;
				}
			}
			if (!remove) {
				sanitiziedClusters.addCluster(cluster);
				sanitiziedClusters.setInputs(cluster, clusters.getInputs(cluster));
				sanitiziedClusters.setOutputs(cluster, clusters.getOutputs(cluster));
			}
		}
		return sanitiziedClusters;
	}
}
